﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Runtime.Serialization.Json;
using System.Text;
using System.Threading.Tasks;
using System.Web.Script.Serialization;
using pmOne.PowerBI.PowerBIObjects;
using Microsoft.IdentityModel.Clients.ActiveDirectory;

namespace pmOne.PowerBI
{
    public class PowerBIClient
    {
        private static List<string> Beta_APIs = new List<string>(){ "reports", "dashboards", "tiles" };

        //Step 2 - Replace redirectUri with the uri you used when you registered your app
        public string RedirectUri = "https://login.live.com/oauth20_desktop.srf";

        //Power BI resource uri
        public const string ResourceUri = "https://analysis.windows.net/powerbi/api";
        //OAuth2 authority
        public const string AuthorityUri = "https://login.windows.net/common/oauth2/authorize";

        //Base API URL
        private const string PBIApiUrl = "https://api.powerbi.com/v1.0/myorg";
        private const string PBIApiUrl_BETA = "https://api.powerbi.com/beta/myorg";

        private string _clientID;
        private AuthenticationContext _authenticationContext = null;
        private string _accessToken = String.Empty;

        DataContractJsonSerializer _serializer;
        NumberStyles _style;
        CultureInfo _culture;

        public PowerBIClient(string clientID)
        {
            _clientID = clientID;

            _style = NumberStyles.Number;
            _culture = CultureInfo.CreateSpecificCulture("en-US");

            Connect();
        }

        public PowerBIClient(string clientID, string accessToken)
        {
            _clientID = clientID;
            _accessToken = accessToken;

            _style = NumberStyles.Number;
            _culture = CultureInfo.CreateSpecificCulture("en-US");
        }

        public PowerBIClient(string clientID, string userName, string password)
        {
            _clientID = clientID;

            _style = NumberStyles.Number;
            _culture = CultureInfo.CreateSpecificCulture("en-US");

            Connect(userName, password);
        }

        public void Connect(string username, string password)
        {
            //http://blogs.msdn.com/b/tomholl/archive/2014/11/25/unattended-authentication-to-azure-management-apis-with-azure-active-directory.aspx
            UserCredential uc = new UserCredential(username, password);

            if (_accessToken == String.Empty)
            {
                TokenCache TC = new TokenCache();
                _authenticationContext = new AuthenticationContext(AuthorityUri, TC);
                _accessToken = _authenticationContext.AcquireToken(ResourceUri, _clientID, uc).AccessToken.ToString();
            }
            else
            {
                _accessToken = _authenticationContext.AcquireTokenSilent(ResourceUri, _clientID).AccessToken;
            }
        }

        public void Connect()
        {
            //Get access token: 
            // To call a Power BI REST operation, create an instance of AuthenticationContext and call AcquireToken
            // AuthenticationContext is part of the Active Directory Authentication Library NuGet package
            if (_accessToken == String.Empty)
            {
                TokenCache TC = new TokenCache();
                _authenticationContext = new AuthenticationContext(AuthorityUri, TC);
                _accessToken = _authenticationContext.AcquireToken(ResourceUri, _clientID, new Uri(RedirectUri)).AccessToken.ToString();
            }
            else
            {
                _accessToken = _authenticationContext.AcquireTokenSilent(ResourceUri, _clientID).AccessToken;
            }
        }

        private string AccessToken
        {
            get
            {
                if (string.IsNullOrEmpty(_accessToken))
                    Connect();
                return _accessToken;
            }
            set
            {
                _accessToken = value;
            }
        }

        public List<PBIDataset> Datasets
        {
            get
            {
                _serializer = new DataContractJsonSerializer(typeof(PBIDatasets));

                PBIDatasets pbids = (PBIDatasets)_serializer.ReadObject(SendGETRequest(PowerBIClient_API.DataSets).GetResponseStream());

                foreach(PBIDataset ds in pbids.Datasets)
                {
                    ds.ParentPowerBIAPI = this;
                }

                return pbids.Datasets;                
            }
        }

        public PBIDataset GetDatasetByID(Guid id)
        {
            foreach (PBIDataset ds in Datasets)
            {
                if (ds.ID == id)
                    return ds;
            }

            return null;
        }

        public PBIDataset GetDatasetByName(string name)
        {
            foreach (PBIDataset ds in Datasets)
            {
                if (ds.Name == name)
                    return ds;
            }

            return null;
        }

        public PBIReport GetReportByName(string name)
        {
            foreach (PBIReport rep in Reports)
            {
                if (rep.Name == name)
                    return rep;
            }

            return null;
        }

        public PBIDashboard GetDashboardByName(string name)
        {
            foreach (PBIDashboard dash in Dashboards)
            {
                if (dash.DisplayName == name)
                    return dash;
            }

            return null;
        }

        public List<PBIDashboard> Dashboards
        {
            get
            {
                _serializer = new DataContractJsonSerializer(typeof(PBIDashboards));

                PBIDashboards pbidb = (PBIDashboards)_serializer.ReadObject(SendGETRequest(PowerBIClient_API.Dashboards).GetResponseStream());

                foreach (PBIDashboard ds in pbidb.Dashboards)
                {
                    ds.ParentPowerBIAPI = this;
                }

                return pbidb.Dashboards;
            }
        }

        public List<PBIReport> Reports
        {
            get
            {
                _serializer = new DataContractJsonSerializer(typeof(PBIReports));

                PBIReports pbirpts = (PBIReports)_serializer.ReadObject(SendGETRequest(PowerBIClient_API.Reports).GetResponseStream());

                foreach (PBIReport ds in pbirpts.Reports)
                {
                    ds.ParentPowerBIAPI = this;
                }

                return pbirpts.Reports;
            }
        }

        #region Private WebRequest Methods
        private HttpWebResponse SendWebRequest(string api, string method, string body)
        {
            string api_url;

            if (Beta_APIs.Contains(api.Trim('/').Split('/')[0]))
                api_url = PBIApiUrl_BETA;
            else
                api_url = PBIApiUrl;

            HttpWebRequest request = System.Net.WebRequest.Create(string.Format("{0}/{1}", api_url, api.ToString().ToLower())) as System.Net.HttpWebRequest;
            request.KeepAlive = true;
            request.Method = method.ToUpper();
            request.ContentLength = 0;
            request.ContentType = "application/json; charset=utf-8";
            request.Headers.Add("Authorization", String.Format("Bearer {0}", AccessToken));

            if (body != null)
            {
                byte[] byteArray = System.Text.Encoding.UTF8.GetBytes(body);
                request.ContentLength = byteArray.Length;

                //Write JSON byte[] into a Stream
                using (Stream writer = request.GetRequestStream())
                {
                    writer.Write(byteArray, 0, byteArray.Length);
                }
            }
            return (HttpWebResponse)request.GetResponse();
        }
        private HttpWebResponse SendWebRequest(string api, string method)
        {
            return SendWebRequest(api, method, null);
        }
        private async Task<HttpWebResponse> SendWebRequestAsync(string api, string method, string body)
        {
            string api_url;

            if (Beta_APIs.Contains(api.Trim('/').Split('/')[0]))
                api_url = PBIApiUrl_BETA;
            else
                api_url = PBIApiUrl;

            HttpWebRequest request = System.Net.WebRequest.Create(string.Format("{0}/{1}", api_url, api.ToString().ToLower())) as System.Net.HttpWebRequest;
            request.KeepAlive = true;
            request.Method = method.ToUpper();
            request.ContentLength = 0;
            request.ContentType = "application/json; charset=utf-8";
            request.Headers.Add("Authorization", String.Format("Bearer {0}", AccessToken));

            if (body != null)
            {
                byte[] byteArray = System.Text.Encoding.UTF8.GetBytes(body);
                request.ContentLength = byteArray.Length;

                //Write JSON byte[] into a Stream
                using (Stream writer = request.GetRequestStream())
                {
                    writer.Write(byteArray, 0, byteArray.Length);
                }
            }
            return (HttpWebResponse)await request.GetResponseAsync();

        }
        private async Task<HttpWebResponse> SendWebRequestAsync(string api, string method)
        {
            return await SendWebRequestAsync(api, method, null);
        }
        #endregion

        #region POST Requests
        public HttpWebResponse SendPOSTRequest(string api, string json)
        {
            return SendWebRequest(api, "POST", json);
        }
        public HttpWebResponse SendPOSTRequest(PowerBIClient_API api, string json)
        {
            return SendPOSTRequest(api.ToString().ToLower(), json);
        }
        public async Task<HttpWebResponse> SendPOSTRequestAsync(string api, string json)
        {
            return await SendWebRequestAsync(api, "POST", json);
        }
        public async Task<HttpWebResponse> SendPOSTRequestAsync(PowerBIClient_API api, string json)
        {
            return await SendWebRequestAsync(api.ToString().ToLower(), "POST", json);
        }
        #endregion

        #region GET Requests
        public HttpWebResponse SendGETRequest(string api)
        {
            return SendWebRequest(api, "GET");
        }
        public HttpWebResponse SendGETRequest(PowerBIClient_API api)
        {
            return SendGETRequest(api.ToString().ToLower());
        }
        public async Task<HttpWebResponse> SendGETRequestAsync(string api)
        {
            return await SendWebRequestAsync(api, "GET");
        }
        public async Task<HttpWebResponse> SendGETRequestAsync(PowerBIClient_API api)
        {
            return await SendWebRequestAsync(api.ToString().ToLower(), "GET");
        }
        #endregion

        #region PUT Requests
        public HttpWebResponse SendPUTRequest(string api, string json)
        {
            return SendWebRequest(api, "PUT", json);
        }
        public async Task<HttpWebResponse> SendPUTRequestAsync(string api, string json)
        {
            return await SendWebRequestAsync(api, "PUT", json);
        }
        #endregion

        #region DELETE Requests
        public HttpWebResponse SendDELETERequest(string api)
        {
            return SendWebRequest(api, "DELETE");
        }
        public async Task<HttpWebResponse> SendDELETERequestAsync(string api)
        {
            return await SendWebRequestAsync(api, "DELETE");
        }
        #endregion
    }

    public enum PowerBIClient_API
    {
        DataSets,
        Reports,
        Groups,
        Dashboards,
        Tiles
    }

    public enum PowerBIClient_DataType
    {
        PBI_Int64,
        PBI_bool,
        PBI_string,
        PBI_DateTime
    }

    public enum PowerBIClient_DefaultRetentionPolicy
    {
        None,
        basicFIFO
    }
}
